import timeit

import calibrationBase
import numpy as np
from karabo.bound import KARABO_CLASSINFO
from karabo.common.states import State

from . import shmem_utils, utils
from ._version import version as deviceVersion
from .base_correction import BaseCorrection
from .agipd_gpu import AgipdGpuRunner


@KARABO_CLASSINFO("AgipdCorrection", deviceVersion)
class AgipdCorrection(BaseCorrection):
    @staticmethod
    def expectedParameters(expected):
        AgipdCorrection.addConstant(
            "ThresholdsDark",
            "Dark",
            expected,
            optional=False,
            mandatoryForIteration=True,
        )
        AgipdCorrection.addConstant(
            "Offset", "Dark", expected, optional=False, mandatoryForIteration=True
        )
        AgipdCorrection.addConstant(
            "SlopesPC", "Dark", expected, optional=True, mandatoryForIteration=True
        )
        AgipdCorrection.addConstant(
            "SlopesFF",
            "Illuminated",
            expected,
            optional=True,
            mandatoryForIteration=True,
        )
        AgipdCorrection.addConstant(
            "BadPixelsDark", "Dark", expected, optional=True, mandatoryForIteration=True
        )
        AgipdCorrection.addConstant(
            "BadPixelsPC", "Dark", expected, optional=True, mandatoryForIteration=True
        )
        AgipdCorrection.addConstant(
            "BadPixelsFF",
            "Illuminated",
            expected,
            optional=True,
            mandatoryForIteration=True,
        )
        super(AgipdCorrection, AgipdCorrection).expectedParameters(expected)

    def __init__(self, config):
        super().__init__(config)
        output_axis_order = config.get("dataFormat.outputAxisOrder")
        if output_axis_order == "pixels-fast":
            self._output_transpose = (0, 2, 1)
        elif output_axis_order == "memorycells-fast":
            self._output_transpose = (2, 1, 0)
        else:
            self._output_transpose = None
        self._offset_map = None
        self._update_pulse_filter(config.get("pulseFilter"))
        self._update_shapes(
            config.get("dataFormat.pixelsX"),
            config.get("dataFormat.pixelsY"),
            config.get("dataFormat.memoryCells"),
            self.pulse_filter,
            self._output_transpose,
        )

        self.updateState(State.ON)

    def process_input(self, data, metadata):
        """Registered for dataInput, handles all processing and sending"""

        if not self.get("doAnything"):
            if self.get("state") is State.PROCESSING:
                self.updateState(State.ACTIVE)
            return

        source = metadata.get("source")

        if source not in self.sources:
            self.log.INFO(f"Ignoring unknown source {source}")
            return

        # TODO: what are these empty things for?
        if not data.has("image"):
            self.log.INFO("Ignoring hash without image node")
            return

        time_start = timeit.default_timer()

        train_id = metadata.getAttribute("timestamp", "tid")
        cell_table = np.squeeze(data.get("image.cellId"))
        assert isinstance(cell_table, np.ndarray), "image.cellId should be ndarray"
        if len(cell_table.shape) == 0:
            msg = "cellId had 0 dimensions. DAQ may not be sending data."
            self.set("status", msg)
            self.log.WARN(msg)
            return
        # original shape: memory_cell, data/raw_gain, x, y
        # TODO: consider making paths configurable
        image_data = data.get("image.data")
        self.log.INFO(f"Image data had shape: {image_data.shape}")
        return
        if image_data.shape[0] != self.get("dataFormat.memoryCells"):
            self.set(
                "status", f"Updating input shapes based on received {image_data.shape}"
            )
            # TODO: truncate if > 800
            self.set("dataFormat.memoryCells", image_data.shape[0])
            with self._buffer_lock:
                self._update_pulse_filter(self.get("pulseFilter"))
                self._update_shapes(
                    self.get("dataFormat.pixelsX"),
                    self.get("dataFormat.pixelsY"),
                    self.get("dataFormat.memoryCells"),
                    self.pulse_filter,
                    self._output_transpose,
                )
        # TODO: check shape (DAQ fake data and RunToPipe don't agree)
        # TODO: consider just updating shapes based on whatever comes in

        correction_cell_num = self.get("dataFormat.memoryCellsCorrection")
        do_generate_preview = train_id % self.get(
            "preview.trainIdModulo"
        ) == 0 and self.get("preview.enable")
        can_apply_correction = correction_cell_num > 0
        do_apply_correction = self.get("applyCorrection")

        if not self.get("state") is State.PROCESSING:
            self.updateState(State.PROCESSING)
            self.set("status", "Processing data")
        if self._state_reset_timer is None:
            self._state_reset_timer = utils.DelayableTimer(
                timeout=self.get("processingStateTimeout"),
                callback=self._reset_state_from_processing,
            )
        else:
            self._state_reset_timer.set_timeout(self.get("processingStateTimeout"))

        with self._buffer_lock:
            cell_table = cell_table[self.pulse_filter]
            pulse_table = np.squeeze(data.get("image.pulseId"))[self.pulse_filter]

            cell_table_max = np.max(cell_table)
            if do_apply_correction:
                if not can_apply_correction:
                    msg = "No constant loaded, correction will not be applied."
                    self.log.WARN(msg)
                    self.set("status", msg)
                    do_apply_correction = False
                elif cell_table_max >= correction_cell_num:
                    msg = (
                        f"Max cell ID ({cell_table_max}) exceeds range for loaded "
                        f"constant (has {correction_cell_num} cells). Some frames "
                        "will not be corrected."
                    )
                    self.log.WARN(msg)
                    self.set("status", msg)

            self.gpu_runner.load_data(image_data)
            buffer_handle, buffer_array = self._shmem_buffer.next_slot()
            if do_apply_correction:
                self.gpu_runner.load_cell_table(cell_table)
                self.gpu_runner.correct()
            else:
                self.gpu_runner.only_cast()
            self.gpu_runner.reshape(out=buffer_array)
            if do_generate_preview:
                preview_slice_index = self.get("preview.pulse")
                if preview_slice_index >= 0:
                    # look at pulse_table to find which index this pulse ID is in
                    pulse_id_found = np.where(pulse_table == preview_slice_index)[0]
                    if len(pulse_id_found) == 0:
                        pulse_found_instead = pulse_table[0]
                        msg = (
                            f"Pulse {preview_slice_index} not found in "
                            f"image.pulseId, arbitrary pulse "
                            f"{pulse_found_instead} will be shown."
                        )
                        preview_slice_index = 0
                        self.log.WARN(msg)
                        self.set("status", msg)
                    else:
                        preview_slice_index = pulse_id_found[0]
                if not do_apply_correction:
                    if can_apply_correction:
                        # in this case, cell table has not been loaded, but needs to be now
                        self.gpu_runner.load_cell_table(cell_table)
                    else:
                        # in this case, there will be no corrected preview
                        self.log.WARN(
                            "Corrected preview will not actually be corrected."
                        )
                preview_raw, preview_corrected = self.gpu_runner.compute_preview(
                    preview_slice_index,
                    have_corrected=do_apply_correction,
                    can_correct=can_apply_correction,
                )

        data.set("image.data", buffer_handle)
        data.set("image.cellId", cell_table[:, np.newaxis])
        data.set("image.pulseId", pulse_table[:, np.newaxis])
        data.set("calngShmemPaths", ["image.data"])
        self._write_output(data, metadata)
        if do_generate_preview:
            self._write_combiner_preview(
                preview_raw, preview_corrected, train_id, source
            )

        # update rate etc.
        self._buffered_status_update.set("trainId", train_id)
        self._rate_tracker.update()
        time_spent = timeit.default_timer() - time_start
        self._buffered_status_update.set(
            "performance.lastProcessingDuration", time_spent * 1000
        )
        if self.get("performance.rateUpdateOnEachInput"):
            self._update_actual_rate()

    def constantLoaded(self):
        """Hook from CalibrationReceiverBaseDevice called after each getConstant

        Here, used to load the received constants (or correction maps derived
        fromt them) onto GPU.

        TODO: call after receiving *all* constants instead of calling once per
        new constant (will cause some overhead for bigger devices)

        """

        self.log.WARN("Not ready to handle constants yet")
        ...

    def _update_pulse_filter(self, filter_string):
        """Called whenever the pulse filter changes, typically followed by
        _update_shapes"""

        if filter_string.strip() == "":
            new_filter = np.arange(self.get("dataFormat.memoryCells"), dtype=np.uint16)
        else:
            new_filter = np.array(eval(filter_string), dtype=np.uint16)
        assert np.max(new_filter) < self.get("dataFormat.memoryCells")
        self.pulse_filter = new_filter

    def _update_shapes(
        self, pixels_x, pixels_y, memory_cells, pulse_filter, output_transpose
    ):
        """(Re)initialize (GPU) buffers according to expected data shapes"""

        input_data_shape = (memory_cells, 1, pixels_y, pixels_x)
        # reflect the axis reordering in the expected output shape
        output_data_shape = utils.shape_after_transpose(
            input_data_shape, output_transpose
        )
        self.set("dataFormat.inputDataShape", list(input_data_shape))
        self.set("dataFormat.outputDataShape", list(output_data_shape))

        if self._shmem_buffer is None:
            shmem_buffer_name = self.getInstanceId() + ":dataOutput"
            memory_budget = self.get("outputShmemBufferSize") * 2 ** 30
            self.log.INFO(f"Opening new shmem buffer: {shmem_buffer_name}")
            self._shmem_buffer = shmem_utils.ShmemCircularBuffer(
                memory_budget,
                output_data_shape,
                self.output_data_dtype,
                shmem_buffer_name,
            )
        else:
            self._shmem_buffer.change_shape(output_data_shape)

        self.gpu_runner = AgipdGpuRunner(
            pixels_x,
            pixels_y,
            memory_cells,
            output_transpose=output_transpose,
            input_data_dtype=self.input_data_dtype,
            output_data_dtype=self.output_data_dtype,
        )

        self._update_maps_on_gpu()

    def _update_maps_on_gpu(self):
        """Updates the correction maps stored on GPU based on constants known

        This only does something useful if constants have been retrieved from
        CalCat.  Should be called automatically upon retrieval and after
        changing the data shape.

        """

        self.set("status", "Updating constants on GPU using known constants")
        self.updateState(State.CHANGING)
        if self._offset_map is not None:
            self.gpu_runner.load_constants(self._offset_map)
            msg = "Done transferring known constant(s) to GPU"
            self.log.INFO(msg)
            self.set("status", msg)

        self.updateState(State.ON)
