import enum

import cupy
import numpy as np

from . import base_gpu, utils


class CorrectionFlags(enum.IntFlag):
    THRESHOLD = 1
    OFFSET = 2
    BLSHIFT = 4
    REL_GAIN_PC = 8
    REL_GAIN_XRAY = 16
    BPMASK = 32


class AgipdGpuRunner(base_gpu.BaseGpuRunner):
    _kernel_source_filename = "agipd_gpu_kernels.cpp"

    def __init__(
        self,
        pixels_x,
        pixels_y,
        memory_cells,
        output_transpose=(1, 2, 0),  # default: memorycells-fast
        constant_memory_cells=None,
        input_data_dtype=np.uint16,
        output_data_dtype=np.float32,
    ):
        self.input_shape = (memory_cells, 2, pixels_x, pixels_y)
        self.processed_shape = (memory_cells, pixels_x, pixels_y)
        super().__init__(
            pixels_x,
            pixels_y,
            memory_cells,
            output_transpose,
            constant_memory_cells,
            input_data_dtype,
            output_data_dtype,
        )
        self.gain_map_gpu = cupy.empty(self.processed_shape, dtype=np.uint8)

        self.map_shape = (self.constant_memory_cells, self.pixels_x, self.pixels_y)
        self.gm_map_shape = self.map_shape + (3,)  # for gain-mapped constants
        self.threshold_map_shape = self.map_shape + (2,)
        # constants
        self.gain_thresholds_gpu = cupy.empty(
            self.threshold_map_shape, dtype=np.float32
        )
        self.offset_map_gpu = cupy.zeros(self.gm_map_shape, dtype=np.float32)
        self.rel_gain_pc_map_gpu = cupy.ones(self.gm_map_shape, dtype=np.float32)
        self.md_additional_offset_gpu = cupy.zeros(self.map_shape, dtype=np.float32)
        self.rel_gain_xray_map_gpu = cupy.ones(self.map_shape, dtype=np.float32)
        self.badpixel_map_gpu = cupy.zeros(self.map_shape, dtype=np.uint32)

        self.update_block_size((1, 1, 64))

    def load_thresholds(self, threshold_map):
        # shape: y, x, memory cell, threshold 0 / threshold 1 / 3 gain values
        # TODO: do we need the gain values (in the constant) for anything?
        self.gain_thresholds_gpu.set(
            np.transpose(threshold_map[..., :2], (2, 1, 0, 3)).astype(np.float32)
        )

    def load_offset_map(self, offset_map):
        # shape: y, x, memory cell, gain stage
        self.offset_map_gpu.set(
            np.transpose(offset_map, (2, 1, 0, 3)).astype(np.float32)
        )

    def load_rel_gain_pc_map(self, slopes_pc_map):
        # pc has funny shape (11, 352, 128, 512) from file
        # this is (fi, memory cell, y, x)
        slopes_pc_map = slopes_pc_map.astype(np.float32)
        pc_high_m = slopes_pc_map[0]
        pc_high_I = slopes_pc_map[1]
        pc_med_m = slopes_pc_map[3]
        pc_med_I = slopes_pc_map[4]
        frac_high_med = pc_high_m / pc_med_m
        # TODO: verify formula
        md_additional_offset = (pc_high_I - pc_med_I * frac_high_med).astype(np.float32)
        rel_gain_map = np.ones(
            (3, self.constant_memory_cells, self.pixels_y, self.pixels_x), dtype=np.float32
        )
        rel_gain_map[1] = rel_gain_map[0] * frac_high_med
        rel_gain_map[2] = rel_gain_map[1] * 4.48
        # TODO: enable overriding this based on user input
        self.rel_gain_pc_map_gpu.set(np.transpose(rel_gain_map, (1, 3, 2, 0)))
        self.md_additional_offset_gpu.set(np.transpose(md_additional_offset, (0, 2, 1)))

    def load_rel_gain_ff_map(self, slopes_ff_map):
        # constant shape: y, x, memory cell
        if slopes_ff_map.shape[2] == 2:
            # old format, is per pixel only
            raise NotImplementedError("Old slopes FF map format")
        self.rel_gain_xray_map_gpu.set(
            np.transpose(slopes_ff_map).astype(np.float32)
        )

    def correct(self, flags):
        """Apply corrections to data (must load constant, data, and cell_table first)

        Applies corrections to input data and casts to desired output dtype.
        Parameter cell_table allows out of order or non-contiguous memory cells
        in input data.  Both input ndarrays are assumed to be on GPU already,
        preferably wrapped in GPU arrays (cupy array).

        Will return string encoded handle to shared memory output buffer and
        (view of) said buffer as an ndarray.  Keep in mind that the output
        buffers will get overwritten eventually (circular buffer).
        """
        if flags & CorrectionFlags.BLSHIFT:
            raise NotImplementedError("Baseline shift not implemented yet")
        self.correction_kernel(
            self.full_grid,
            self.full_block,
            (
                self.input_data_gpu,
                self.cell_table_gpu,
                np.uint8(flags),
                self.gain_thresholds_gpu,
                self.offset_map_gpu,
                self.rel_gain_pc_map_gpu,
                self.md_additional_offset_gpu,
                self.rel_gain_xray_map_gpu,
                self.badpixel_map_gpu,
                self.gain_map_gpu,
                self.processed_data_gpu,
            ),
        )

    def _init_kernels(self):
        kernel_source = self._kernel_template.render(
            {
                "pixels_x": self.pixels_x,
                "pixels_y": self.pixels_y,
                "data_memory_cells": self.memory_cells,
                "constant_memory_cells": self.constant_memory_cells,
                "input_data_dtype": utils.np_dtype_to_c_type(self.input_data_dtype),
                "output_data_dtype": utils.np_dtype_to_c_type(self.output_data_dtype),
            }
        )
        self.source_module = cupy.RawModule(code=kernel_source)
        self.correction_kernel = self.source_module.get_function("correct")
        self.casting_kernel = self.source_module.get_function("only_cast")
