import cupy
import numpy as np

from . import base_gpu, utils


class DsscGpuRunner(base_gpu.BaseGpuRunner):
    _kernel_source_filename = "dssc_gpu_kernels.cpp"

    def __init__(
        self,
        pixels_x,
        pixels_y,
        memory_cells,
        output_transpose=(2, 1, 0),  # default: memorycells-fast
        constant_memory_cells=None,
        input_data_dtype=np.uint16,
        output_data_dtype=np.float32,
    ):
        self.input_shape = (memory_cells, pixels_y, pixels_x)
        self.processed_shape = self.input_shape
        super().__init__(
            pixels_x,
            pixels_y,
            memory_cells,
            output_transpose,
            constant_memory_cells,
            input_data_dtype,
            output_data_dtype,
        )
        self.map_shape = (self.pixels_x, self.pixels_y, self.constant_memory_cells)
        self.offset_map_gpu = cupy.empty(self.map_shape, dtype=np.float32)

        self._init_kernels()

        self.offset_map_gpu = cupy.empty(self.map_shape, dtype=np.float32)

        self.update_block_size((1, 1, 64))

    def load_constants(self, offset_map):
        constant_memory_cells = offset_map.shape[-1]
        if constant_memory_cells != self.constant_memory_cells:
            self.constant_memory_cells = constant_memory_cells
            self.map_shape = (self.pixels_x, self.pixels_y, self.constant_memory_cells)
            self.offset_map_gpu = cupy.empty(self.map_shape, dtype=np.float32)
            self._init_kernels()
        self.offset_map_gpu.set(offset_map)

    def correct(self):
        """Apply corrections to data (must load constant, data, and cell_table first)

        Applies corrections to input data and casts to desired output dtype.
        Parameter cell_table allows out of order or non-contiguous memory cells
        in input data.  Both input ndarrays are assumed to be on GPU already,
        preferably wrapped in GPU arrays (cupy array).

        Will return string encoded handle to shared memory output buffer and
        (view of) said buffer as an ndarray.  Keep in mind that the output
        buffers will get overwritten eventually (circular buffer).
        """
        self.correction_kernel(
            self.full_grid,
            self.full_block,
            (
                self.input_data_gpu,
                self.cell_table_gpu,
                self.offset_map_gpu,
                self.processed_data_gpu,
            ),
        )

    def _init_kernels(self):
        kernel_source = self._kernel_template.render(
            {
                "pixels_x": self.pixels_x,
                "pixels_y": self.pixels_y,
                "data_memory_cells": self.memory_cells,
                "constant_memory_cells": self.constant_memory_cells,
                "input_data_dtype": utils.np_dtype_to_c_type(self.input_data_dtype),
                "output_data_dtype": utils.np_dtype_to_c_type(self.output_data_dtype),
            }
        )
        self.source_module = cupy.RawModule(code=kernel_source)
        self.correction_kernel = self.source_module.get_function("correct")
        self.casting_kernel = self.source_module.get_function("only_cast")
